import random
import weakref

from util import alias_match

class Activity(object):
    __slots__ = ('id', 'room', 'properties', 'members')

    def __init__(self, id, room, properties=None, members=None):
        self.id = id
        self.room = room

        if properties:
            self.properties = properties
        else:
            self.properties = {}

        if members:
            self.members = members
        else:
            self.members = set()

    def __repr__(self):
        return '<Activity %r %r>' % (self.id, self.room)

    def __cmp__(self, other):
        return cmp(self.id, other.id)

    def __eq__(self, other):
        return ((self.id, self.room, self.properties, self.members) ==
                (other.id, other.room, other.properties, other.members))

    def __getstate__(self):
        return (self.id, self.room, self.properties, self.members)

    def __setstate__(self, (id, room, properties, members)):
        self.id = id
        self.room = room
        self.properties = properties
        self.members = members

class Buddy(object):
    __slots__ = ('jid', 'properties', 'current_activity', 'alias')

    def __init__(self, jid, properties=None):
        self.jid = jid

        if properties:
            self.properties = properties
        else:
            self.properties = {}

        # (room, activity id)
        self.current_activity = None

        self.alias = None

    def __repr__(self):
        return '<Buddy %r (%r) %r>' % (self.jid, self.alias, self.properties)

    def __cmp__(self, other):
        return cmp(self.jid, other.jid)

    def __eq__(self, other):
        return (self.jid, self.properties, self.current_activity, self.alias) == \
                (other.jid, other.properties, other.current_activity, other.alias)

    def __getstate__(self):
        return (self.jid, self.properties, self.current_activity, self.alias)

    def __setstate__(self, (jid, properties, current_activity, alias)):
        self.jid = jid
        self.properties = properties
        self.current_activity = current_activity
        self.alias = alias

class View(object):
    def __init__(self, model, size, test, handler):
        self.model = model
        self.size = size
        self._test = test
        self.handler = handler
        self.results = []

    def test(self, obj):
        return self._test(obj)

    def __iter__(self):
        return iter(self.results)

class ActivityView(View):
    def add_activity(self, activity):
        if len(self.results) >= self.size:
            return

        self.results.append(activity)
        self.handler.activity_found(activity)

    def remove_activity(self, activity):
        self.results.remove(activity)
        self.handler.activity_gone(activity)

class BuddyView(View):
    def add_buddy(self, buddy):
        if len(self.results) >= self.size:
            return

        self.results.append(buddy)
        self.handler.buddy_added(buddy)

    def remove_buddy(self, buddy):
        self.results.remove(buddy)
        self.handler.buddy_removed(buddy)

class GadgetModel(object):
    def __init__(self):
        self.activities = {}
        self.buddies = {}
        # Store views in a weak dict so they are automatically removed when
        # the component close them.
        # We don't use the associated value.
        self.activity_views = weakref.WeakKeyDictionary()
        self.buddy_views = weakref.WeakKeyDictionary()

    def __getstate__(self):
        return (
            self.activities,
            self.buddies,
            list(self.activity_views),
            list(self.buddy_views))

    def __setstate__(self, state):
        (activities, buddies, activity_views, buddy_views) = state
        self.activities = activities
        self.buddies = buddies
        self.activity_views = \
            weakref.WeakKeyDictionary.fromkeys(activity_views, True)
        self.buddy_views = \
            weakref.WeakKeyDictionary.fromkeys(buddy_views, True)

    def _property_match(self, obj, properties):
        "Check: for (k, v) in properties, (k, v) is in obj.properties."

        for key, (type, value) in properties.iteritems():
            if obj.properties.get(key) != (type, value):
                return False

        return True

    def add_activity(self, activity):
        self.activities[activity.id] = activity

        for view in self.activity_views:
            if view.test(activity):
                view.add_activity(activity)

    def create_activity(self, id, room, properties):
        activity = Activity(id, room, properties)
        self.add_activity(activity)
        return activity

    def activity_update(self, id, properties):
        activity = self.activities[id]
        activity.properties = properties

        self.update_activity_views(activity)

    def update_activity_views(self, activity):
        for view in self.activity_views:
            in_ = activity in view.results
            test = view.test(activity)

            if in_ and not test:
                view.remove_activity(activity)

                # Activity has been removed so maybe we could
                # add others activities to complete the view
                self._complete_activity_view(view)
            elif test and not in_:
                view.add_activity(activity)

    def activity_remove(self, id):
        activity = self.activities.pop(id)

        for view in self.activity_views:
            # Use explicit loop with 'is' comparison because it's faster than
            # using 'in', which using ==.

            for other in view.results:
                if activity is other:
                    view.remove_activity(activity)

                    # Activity has been removed so maybe we could
                    # add others activities to complete the view
                    self._complete_activity_view(view)

    def activity_by_id(self, id):
        activity = self.activities[id]
        return activity

    def activity_by_room(self, room):
        # FIXME: There should be a 1-1 mapping of rooms to activities.
        results = []

        for activity in self.activities.itervalues():
            if activity.room == room:
                return activity

        return None

    def activity_by_participants(self, buddies):
        # FIXME: this is not very efficient
        results = []

        for activity in self.activities.itervalues():
            if set(buddies).issubset(activity.members):
                    results.append(activity)

        return results

    def activity_by_properties(self, properties):
        results = []

        for activity in self.activities.itervalues():
            if self._property_match(activity, properties):
                results.append(activity)

        return results

    def activity_all(self):
        return self.activities.values()

    def activity_search(self, criteria):
        found = None

        room = criteria.get('room')
        properties = criteria.get('properties')
        participants = criteria.get('participants')

        if criteria == {}:
            # no criteria return all the activities
            return self.activity_all()

        if room is not None:
            activity = self.activity_by_room(room)

            if activity is not None:
                found = set([activity])

        if properties is not None:
            activities = set(self.activity_by_properties(properties))
            if found is None:
                found = activities
            else:
                found.intersection_update(activities)

        if participants is not None:
            activities = set(self.activity_by_participants(participants))
            if found is None:
                found = activities
            else:
                found.intersection_update(activities)

        if found is not None:
            return found
        else:
            return set()

    def activity_random(self, max_activities):
        if max_activities <= 0:
            raise ValueError("max activities have to be greater than 0")

        if max_activities >= len(self.activities):
            # we want all the activities
            return self.activity_all()

        return random.sample(self.activities.values(), max_activities)

    def activity_view(self, handler, size, criteria):
        properties = criteria.get('properties')
        members = criteria.get('participants')

        def test(activity):
            if properties is not None and \
                not self._property_match(activity, properties):
                    return False

            if members is not None and \
                not self._members_match(activity, members):
                    return False

            return True

        view = ActivityView(self, size, test, handler)
        self.activity_views[view] = True
        results = []

        for activity in self.activities.itervalues():
            if len(results) >= view.size:
                break

            if view.test(activity):
                results.append(activity)

        view.results = results
        return view

    def add_buddy(self, buddy):
        self.buddies[buddy.jid] = buddy

        for view in self.buddy_views:
            if view.test(buddy):
                view.add_buddy(buddy)

    def create_buddy(self, jid, properties):
        buddy = Buddy(jid, properties)
        self.add_buddy(buddy)
        return buddy

    def buddy_view(self, handler, size, criteria):
        alias = criteria.get('alias')
        properties = criteria.get('properties')

        def test(buddy):
            if alias is not None:
                # have to check alias
                if buddy.alias is None or not alias_match(buddy.alias, alias):
                    return False

            if properties is not None and \
                not self._property_match(buddy, properties):
                    return False

            return True

        view = BuddyView(self, size, test, handler)
        self.buddy_views[view] = True
        results = []

        for buddy in self.buddies.itervalues():
            if len(results) >= view.size:
                break

            if view.test(buddy):
                results.append(buddy)

        view.results = results
        return view

    def buddy_update(self, jid, properties):
        buddy = self.buddies[jid]
        buddy.properties = properties

        self.update_buddy_views(buddy)

    def _complete_buddy_view(self, view):
        for buddy in self.buddies.itervalues():
            if len(view.results) >= view.size:
                return

            # Is buddy already in the view?
            for other in view.results:
                if buddy is other:
                    break
            else:
                # buddy is not in the view
                if view.test(buddy):
                    view.add_buddy(buddy)

    def _complete_activity_view(self, view):
        for activity in self.activities.itervalues():
            if len(view.results) >= view.size:
                return

            # Is activity already in the view?
            for other in view.results:
                if activity is other:
                    break
            else:
                # activity is not in the view
                if view.test(activity):
                    view.add_activity(activity)

    def update_buddy_views(self, buddy):
        for view in self.buddy_views:
            in_ = buddy in view.results
            test = view.test(buddy)

            if in_:
                if not test:
                    view.remove_buddy(buddy)

                    # Buddy has been removed so maybe we could
                    # add others buddies to complete the view
                    self._complete_buddy_view(view)
                else:
                    # notify about the change
                    view.handler.buddy_changed(buddy)
            elif test and not in_:
                view.add_buddy(buddy)

        # FIXME: this is not very efficient
        for view in self.activity_views:
            for activity in view.results:
                if buddy.jid in activity.members:
                    view.handler.buddy_changed(buddy)

    def buddy_change_current_activity(self, jid, current_activity):
        buddy = self.buddies[jid]
        buddy.current_activity = current_activity

        for view in self.buddy_views:
            if buddy in view.results:
                # notify about the change
                view.handler.buddy_current_activity_changed(buddy)

        # FIXME: this is not very efficient
        for view in self.activity_views:
            for activity in view.results:
                if buddy.jid in activity.members:
                    view.handler.buddy_current_activity_changed(buddy)

    def buddy_remove(self, jid):
        buddy = self.buddies.pop(jid)

        for view in self.buddy_views:
            for other in view.results:
                if buddy is other:
                    view.remove_buddy(buddy)

                    # Buddies have been removed so maybe we could
                    # add others buddies to complete the view
                    self._complete_buddy_view(view)
                    break

    def buddy_by_id(self, jid):
        buddy = self.buddies[jid]
        return buddy

    def buddy_by_properties(self, properties):
        results = []

        for buddy in self.buddies.itervalues():
            if self._property_match(buddy, properties):
                results.append(buddy)

        return results

    def buddy_all(self):
        return self.buddies.values()

    def buddy_search(self, criteria):
        found = None

        if criteria == {}:
            # no criteria return all the buddies
            return self.buddy_all()

        jid = criteria.get('jid')
        properties = criteria.get('properties')

        if jid is not None:
            buddy = self.buddies.get(jid)
            if buddy is not None:
                found = set([buddy])

        if properties is not None:
            buddies = set(self.buddy_by_properties(properties))
            if found is None:
                found = buddies
            else:
                found.intersection_update(buddies)

        if found is not None:
            return found
        else:
            return set()

    def buddy_random(self, max_buddies):
        if max_buddies <= 0:
            raise ValueError("max buddies have to be greater than 0")

        if max_buddies >= len(self.buddies):
            # we want all the buddies
            return self.buddy_all()

        return random.sample(self.buddies.values(), max_buddies)

    def has_buddy(self, jid):
        return jid in self.buddies

    def has_activity(self, id):
        return id in self.activities

    def activity_add_member(self, id, member):
        activity = self.activities[id]
        if member in activity.members:
            return

        activity.members.add(member)

        for view in self.activity_views:
            for other in view.results:
                if activity is other:
                    view.handler.member_added(activity, member)

        self.update_activity_views(activity)

    def activity_remove_member(self, id, member):
        activity = self.activities[id]
        if member not in activity.members:
            return

        activity.members.remove(member)

        self.update_activity_views(activity)

        for view in self.activity_views:
            for other in view.results:
                if activity is other:
                    view.handler.member_removed(activity, member)

    def buddy_change_alias(self, jid, alias):
        buddy = self.buddies[jid]

        if alias == '':
            alias = None
        buddy.alias = alias

        self.update_buddy_views(buddy)

    def _members_match(self, activity, members):
        return set(members).issubset(activity.members)

