
import unittest
from twisted.internet import defer, reactor
from twisted.protocols import loopback
from twisted.trial.unittest import TestCase as TrialTestCase
from twisted.words.protocols.jabber import component, xmlstream
from twisted.words.xish import domish
from twisted.words.protocols.jabber.client import IQ

from component import (
    GadgetService, PropertyTypeError, parse_properties, properties_to_xml)
from model import GadgetModel, Activity
from util import elem_iq, elem, make_iq_result, xpath_query

import ns

class TestAuthenticator(xmlstream.Authenticator):
    def streamStarted(self, root=None):
        self.xmlstream.sendHeader()
        self.xmlstream.addOnetimeObserver('//handshake', self.handshake)

    def handshake(self, stanza):
        handshake = domish.Element((None, 'handshake'))
        self.xmlstream.send(handshake)
        self.xmlstream.dispatch(self.xmlstream, xmlstream.STREAM_AUTHD_EVENT)

class TestProtocol(xmlstream.XmlStream):
    initiating = False
    namespace = 'jabber:client'

    def __init__(self):
        xmlstream.XmlStream.__init__(self, TestAuthenticator())
        self.addOnetimeObserver(xmlstream.STREAM_AUTHD_EVENT,
            self.authenticated)
        self.ready = defer.Deferred()

    def authenticated(self, event):
        self.ready.callback(None)

class TestCase(TrialTestCase):
    def setUp(self):
        # Ew.
        self.server = TestProtocol()
        auth = component.ConnectComponentAuthenticator('gadget.localhost',
                'bar')
        self.client = xmlstream.XmlStream(auth)
        self.model = GadgetModel()

        # No need to create a GadgetRoster object. set()'s API
        # is enough
        self.roster = set()

        self.service = GadgetService(self.model, self.roster)
        self.service.debug = True
        self.service.parent = self.client
        self.service.parent.jabberId = "gadget.localhost"
        loopback.loopbackAsync(self.server, self.client)
        self.service.componentConnected(self.client)
        return self.server.ready

    def tearDown(self):
        self.client.transport.loseConnection()
        self.server.transport.loseConnection()

    def send_muc_presence(self, _from, jid=None, type=None):
        presence = domish.Element((ns.CLIENT, 'presence'))
        presence['from'] = _from
        x = presence.addElement((ns.MUC_USER, 'x'))

        if type is not None:
            presence['type'] = type

        if jid is not None:
            item = x.addElement('item')
            item['jid'] = jid

        self.server.send(presence)

class UnknownIqTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        iq.send()

        return self.done

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'error'
        error = stanza.firstChildElement()
        assert error.name == 'error'
        br = error.firstChildElement()
        assert br.name == 'feature-not-implemented'
        self.done.callback(None)

class DiscoInfoTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.DISCO_INFO, 'query'))
        iq.send()

        return self.done

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'result'
        query = stanza.firstChildElement()
        names = [x.name for x in query.elements()]
        assert names == ['identity', 'feature', 'feature', 'feature', 'feature', 'feature']
        self.done.callback(None)

class SubscriptionTest(TestCase):
    """Test if Gadget accept subscription requests from users, add them to the model
    and request their properties."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence1)

        presence = domish.Element((ns.CLIENT, 'presence'))
        presence['from'] = 'bob@foo.org'
        presence['to'] = 'gadget.foo.org'
        presence['type'] = 'subscribe'
        self.server.send(presence)
        return self.done

    def presence1(self, stanza):
        assert stanza['type'] == 'subscribed'
        self.server.addOnetimeObserver('//presence', self.presence2)

    def presence2(self, stanza):
        assert stanza['type'] == 'subscribe'
        self.server.addObserver('//iq', self.iq)
        self.waiting = ['properties', 'current_act', 'nick']
        self.server.addOnetimeObserver('//presence', self.presence3)

        presence = domish.Element((ns.CLIENT, 'presence'))
        presence['from'] = 'bob@foo.org'
        presence['to'] = 'gadget.foo.org'
        presence['type'] = 'subscribed'
        self.server.send(presence)

    def presence3(self, stanza):
        # Gadget sent its presence
        # Buddy has been added to the model
        buddy = self.service.model.buddy_by_id('bob@foo.org')
        assert buddy.jid == 'bob@foo.org'

        assert self.roster == set(['bob@foo.org'])

    def iq(self, stanza):
        result = make_iq_result(self.server, stanza)
        pubsub = result.firstChildElement()

        items = xpath_query('/iq/pubsub/items', stanza)[0]
        if items['node'] == ns.BUDDY_PROP:
            child = self.iq_buddy_prop(stanza)
        elif items['node'] == ns.CURRENT_ACTIVITY:
            child = self.iq_current_act(stanza)
        elif items['node'] == ns.NICK:
            child = self.iq_nick(stanza)

        pubsub.addChild(child)
        self.server.send(result)

        if len(self.waiting) == 0:
            reactor.callLater(0, self.later)

    def iq_buddy_prop(self, stanza):
        items = elem('items', node=ns.BUDDY_PROP)(
            elem('item', id='foo')(
                elem(ns.BUDDY_PROP, 'properties')(
                    elem('property', type='str', name='color')(u'red'))))

        self.waiting.remove('properties')
        return items

    def iq_current_act(self, stanza):
        items = elem('items', node=ns.CURRENT_ACTIVITY)(
            elem('item', id='foo')(
                elem(ns.CURRENT_ACTIVITY, 'activity', type='activity1', room='myroom')))

        self.waiting.remove('current_act')
        return items

    def iq_nick(self, stanza):
        items = elem('items', node=ns.NICK)(
            elem('item', id='foo')(
                elem(ns.NICK, 'nick')(u'BOB')))

        self.waiting.remove('nick')
        return items

    def later(self):
        # Buddy properties have been updated
        buddy = self.service.model.buddy_by_id('bob@foo.org')
        assert buddy.jid == 'bob@foo.org'
        assert buddy.properties == {'color': ('str', 'red')}
        assert buddy.alias == 'BOB'
        assert buddy.current_activity == Activity('activity1', 'myroom')

        self.done.callback(None)

class ProbePresenceTest(TestCase):
    """test if gadget adds buddies to the model when it receives their
    probe presence"""
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)

        presence = domish.Element((ns.CLIENT, 'presence'))
        presence['from'] = 'bob@foo.org/Resource'
        presence['to'] = 'gadget.foo.org'
        presence['type'] = 'probe'
        self.server.send(presence)
        return self.done

    def presence(self, stanza):
        assert stanza.name == 'presence'

        buddy = self.service.model.buddy_by_id('bob@foo.org')
        assert buddy.jid == 'bob@foo.org'

        self.done.callback(None)

class UnavailablePresenceTest(TestCase):
    """test if gadget removes buddies from the model when it receives their
    unavailable presence"""
    def runTest(self):
        self.done = defer.Deferred()
        self.service.model.create_buddy('bob@foo.org', {})

        presence = domish.Element((ns.CLIENT, 'presence'))
        presence['from'] = 'bob@foo.org'
        presence['to'] = 'gadget.foo.org'
        presence['type'] = 'unavailable'
        self.server.send(presence)

        reactor.callLater(0, self.later)
        return self.done

    def later(self):
        self.assertRaises(KeyError, self.service.model.buddy_by_id,
                'bob@foo.org')
        self.done.callback(None)

class InviteTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)

        # Bob notifies us of activity properties. (Preparatory step for
        # invitation.)
        m = domish.Element((ns.CLIENT, 'message'))
        ps = m.addElement((ns.ACTIVITY_PROP, 'properties'))
        ps['activity'] = '1'
        ps['room'] = 'chat@conf.foo.org'
        p = ps.addElement((None, 'property'))
        p['type'] = 'str'
        p['name'] = 'foo'
        p.addChild('hello')
        self.server.send(m)

        # Bob has invited us to an activity.
        message = domish.Element((None, 'message'))
        message['from'] = 'chat@conf.foo.org'
        message['to'] = 'inspector@gadget.bar.org'
        x = message.addElement((ns.MUC_USER, 'x'))
        invite = x.addElement((None, 'invite'))
        invite['from'] = 'bob@baz.org'
        self.server.send(message)
        return self.done

    def presence(self, stanza):
        assert stanza['to'] == 'chat@conf.foo.org/inspector'
        assert stanza['from'] == 'inspector@gadget.bar.org'
        x = stanza.firstChildElement()
        assert (x.uri, x.name) == (ns.MUC_USER, 'x')
        self.done.callback(None)

class RoomMembershipTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)
        # Simulate invitation process.
        self.service.create_room('chat@conf.foo.org', 'activity1', {})
        self.service.invited('inspector@gadget.bar.org', 'chat@conf.foo.org')
        return self.done

    def presence(self, _):
        self.send_muc_presence('chat@conf.foo.org/amy', 'amy@foo.org/resource')
        self.send_muc_presence('chat@conf.foo.org/bob', 'bob@foo.org/resource')
        # Oscar won't be added as members because his presence doesn't
        # contain his jid
        self.send_muc_presence('chat@conf.foo.org/oscar')
        self.send_muc_presence('chat@conf.foo.org/inspector')

        reactor.callLater(0, self.blah)

    def blah(self):
        activity = self.model.activity_by_room('chat@conf.foo.org')
        assert activity.members == set(['amy@foo.org', 'bob@foo.org'])
        activity = self.model.activity_by_id('activity1')
        assert activity.members == set(['amy@foo.org', 'bob@foo.org'])

        # Amy leaves the room.
        self.send_muc_presence('chat@conf.foo.org/amy', 'amy@foo.org/resource',
                'unavailable')

        # Cat joins the room.
        self.send_muc_presence('chat@conf.foo.org/cat', 'cat@foo.org/resource')

        assert activity.members == set(['bob@foo.org', 'cat@foo.org'])
        activity = self.model.activity_by_id('activity1')
        assert activity.members == set(['bob@foo.org', 'cat@foo.org'])
        self.done.callback(None)

class PropertiesTest(TestCase):
    def runTest(self):
        m = domish.Element((ns.CLIENT, 'message'))
        ps = m.addElement((ns.ACTIVITY_PROP, 'properties'))
        ps['activity'] = '1'
        ps['room'] = 'chat@conf.foo.org'
        p = ps.addElement((None, 'property'))
        p['type'] = 'str'
        p['name'] = 'foo'
        p.addChild('hello')
        self.server.send(m)

        activity = self.service.pending_activities['chat@conf.foo.org']
        assert activity.properties['foo'] == ('str', 'hello')

class ParsePropertiesTest(unittest.TestCase):
    def test_str_property(self):
        properties = domish.Element((None, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'str'
        property['name'] = 'a_str'
        property.addContent('the string')

        props = parse_properties(properties)
        assert props == {'a_str': ('str', 'the string')}

    def test_int_property(self):
        properties = domish.Element((None, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'int'
        property['name'] = 'a_int'
        property.addContent('69')

        props = parse_properties(properties)
        assert props == {'a_int': ('int', '69')}

    def test_uint_property(self):
        properties = domish.Element((None, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'uint'
        property['name'] = 'an_uint'
        property.addContent('7')

        props = parse_properties(properties)
        assert props == {'an_uint': ('uint', '7')}

    def test_bool_property(self):
        properties = domish.Element((None, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'bool'
        property['name'] = 'a_bool'
        property.addContent('true')

        props = parse_properties(properties)
        assert props == {'a_bool': ('bool', 'true')}

    def test_bytes_property(self):
        properties = domish.Element((None, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'bytes'
        property['name'] = 'some_bytes'
        property.addContent('ABCDEF')

        props = parse_properties(properties)
        assert props == {'some_bytes': ('bytes', 'ABCDEF')}

    def test_invalid_bool_property(self):
        properties = domish.Element((None, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'bool'
        property['name'] = 'a_bool'
        property.addContent('crack')
        self.assertRaises(PropertyTypeError, parse_properties, properties)

    def test_invalid_property(self):
        properties = domish.Element((ns.BUDDY_PROP, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'wtf'
        property['name'] = 'crack'
        property.addContent('hey')
        self.assertRaises(PropertyTypeError, parse_properties, properties)

class PropertiesToXmlTest(unittest.TestCase):
    def test_str_property(self):
        props = {'a_str': ('str', 'the string')}
        properties = properties_to_xml(None, props)

        elements = sorted(properties.elements())
        assert len(elements) == 1
        property = elements[0]
        assert property.name == 'property'
        assert property['type'] == 'str'
        assert property['name'] == 'a_str'
        child = property.children[0]
        assert child == 'the string'

    def test_int_property(self):
        props = {'a_int': ('int', '69')}
        properties = properties_to_xml(None, props)

        elements = sorted(properties.elements())
        assert len(elements) == 1
        property = elements[0]
        assert property.name == 'property'
        assert property['type'] == 'int'
        assert property['name'] == 'a_int'
        child = property.children[0]
        assert child == '69'

    def test_uint_property(self):
        props = {'an_uint': ('uint', '7')}
        properties = properties_to_xml(None, props)

        elements = sorted(properties.elements())
        assert len(elements) == 1
        property = elements[0]
        assert property.name == 'property'
        assert property['type'] == 'uint'
        assert property['name'] == 'an_uint'
        child = property.children[0]
        assert child == '7'

    def test_bool_property(self):
        props = {'a_bool': ('bool', 'true')}
        properties = properties_to_xml(None, props)

        elements = sorted(properties.elements())
        assert len(elements) == 1
        property = elements[0]
        assert property.name == 'property'
        assert property['type'] == 'bool'
        assert property['name'] == 'a_bool'
        child = property.children[0]
        assert child == 'true'

    def test_bytes_property(self):
        props = {'some_bytes': ('bytes', 'ABCDEF')}
        properties = properties_to_xml(None, props)

        elements = sorted(properties.elements())
        assert len(elements) == 1
        property = elements[0]
        assert property.name == 'property'
        assert property['type'] == 'bytes'
        assert property['name'] == 'some_bytes'
        child = property.children[0]
        assert child == 'ABCDEF'

class CapsTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)

        presence = domish.Element((ns.CLIENT, 'presence'))
        presence['from'] = 'bob@foo.org'
        presence['to'] = 'gadget.foo.org'
        presence['type'] = 'probe'
        self.server.send(presence)
        return self.done

    def presence(self, stanza):
        assert stanza.name == 'presence'
        c = stanza.firstChildElement()
        assert c.name == 'c'
        assert c.uri == ns.CAPS

        self.server.addOnetimeObserver('//iq', self.iq)
        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.DISCO_INFO, 'query'))
        query['node'] = c['node']
        iq.send()

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'result'
        query = stanza.firstChildElement()
        assert query.name == 'query'
        identities = set()
        features = set()

        for elt in query.children:
            assert elt.name in ('feature', 'identity')
            if elt.name == 'identity':
                identities.add((elt['category'], elt['type'], elt['name']))
            elif elt.name == 'feature':
                features.add(elt['var'])

        assert identities == set([('collaboration', 'gadget', 'OLPC Gadget')])
        assert features == \
            set([ns.ACTIVITY, ns.BUDDY, "%s+notify" % ns.BUDDY_PROP,
                "%s+notify" % ns.CURRENT_ACTIVITY,
                "%s+notify" % ns.NICK])

        self.done.callback(None)

class BuddyPropertiesChangeTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.service.model.create_buddy('bob@foo.org', {})

        message = domish.Element((None, 'message'))
        message['from'] = 'bob@foo.org'
        message['to'] = 'gadget.foo.org'
        event = message.addElement((ns.PUBSUB_EVENT, 'event'))
        items = event.addElement((None, 'items'))
        items['node'] = ns.BUDDY_PROP
        item = items.addElement((None, 'item'))
        properties = item.addElement((ns.BUDDY_PROP, 'properties'))
        property = properties.addElement((None, 'property'))
        property['name'] = 'color'
        property['type'] = 'str'
        property.addContent('red')
        self.server.send(message)

        reactor.callLater(0, self.later1)
        return self.done

    def later1(self):
        buddy = self.service.model.buddy_by_id('bob@foo.org')
        assert buddy.properties == {'color': ('str', 'red')}

        message = domish.Element((None, 'message'))
        message['from'] = 'bob@foo.org'
        message['to'] = 'gadget.foo.org'
        event = message.addElement((ns.PUBSUB_EVENT, 'event'))
        items = event.addElement((None, 'items'))
        items['node'] = ns.BUDDY_PROP
        item = items.addElement((None, 'item'))
        properties = item.addElement((ns.BUDDY_PROP, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'str'
        property['name'] = 'key'
        property.addContent('my key')
        property = properties.addElement((None, 'property'))
        property['type'] = 'str'
        property['name'] = 'color'
        property.addContent('blue')

        self.server.send(message)
        reactor.callLater(0, self.later2)

    def later2(self):
        buddy = self.service.model.buddy_by_id('bob@foo.org')
        assert buddy.properties == {'color': ('str', 'blue'),
            'key': ('str', 'my key')}
        self.done.callback(None)

class ActivityPropertiesTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)
        # Simulate invitation process.
        self.service.create_room('chat@conf.localhost', 'activity1', {})
        self.service.invited('inspector@gadget.localhost',
            'chat@conf.localhost')
        return self.done

    def presence(self, _):
        # Bob's muc presence
        self.send_muc_presence('chat@conf.localhost/bob',
            'bob@foo.org/resource')

        # Gadget's muc presence
        self.send_muc_presence('chat@conf.localhost/inspector')

        reactor.callLater(0, self.later)

    def later(self):
        activity = self.model.activity_by_room('chat@conf.localhost')
        assert activity.members == set(['bob@foo.org'])

        activity = self.service.model.activity_by_id('activity1')
        assert activity == Activity('activity1', 'chat@conf.localhost', {},
                set(['bob@foo.org']))

        # bob changes activity properties
        message = domish.Element((None, 'message'))
        message['type'] = 'groupchat'
        message['to'] = 'chat@conf.localhost'
        message['from'] = 'chat@conf.localhost/bob'
        properties = message.addElement((ns.ACTIVITY_PROP, 'properties'))
        properties['activity'] = 'activity1'
        properties['room'] = 'chat@conf.localhost'
        property = properties.addElement((None, 'property'))
        property['type'] = 'str'
        property['name'] = 'title'
        property.addContent('Test Activity')
        self.server.send(message)

        reactor.callLater(0, self.later2)

    def later2(self):
        activity = self.service.model.activity_by_id('activity1')
        assert activity == Activity('activity1', 'chat@conf.localhost',
                {'title': ('str', 'Test Activity')}, set(['bob@foo.org']))

        self.done.callback(None)

class ActivityLeaveTest(TestCase):
    """Gadget leaves an activity and forgets about it when the last
       buddy leaves."""
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)
        # Simulate invitation process.
        self.service.create_room('chat@conf.localhost', 'activity1', {})
        self.service.invited('inspector@gadget.localhost',
                'chat@conf.localhost')
        return self.done

    def presence(self, _):
        # Georges's muc presence
        self.send_muc_presence('chat@conf.localhost/georges',
                'georges@foo.org/resource')

        # Gadget's muc presence
        self.send_muc_presence('chat@conf.localhost/inspector')

        reactor.callLater(0, self.later)

    def later(self):
        activity = self.service.model.activity_by_id('activity1')
        assert activity == Activity('activity1', 'chat@conf.localhost', {},
                set(['georges@foo.org']))

        self.server.addOnetimeObserver('//presence', self.presence2)
        self.send_muc_presence('chat@conf.localhost/georges',
                'georges@foo.org/resource', 'unavailable')

    def presence2(self, stanza):
        assert stanza.name == 'presence'
        assert stanza['to'] == 'chat@conf.localhost/inspector'
        assert stanza['from'] == 'gadget.localhost'
        assert stanza['type'] == 'unavailable'

        assert not self.model.has_activity('activity1')
        assert len(self.model.activities) == 0
        self.done.callback(None)

class ActivityBecomesPrivate(TestCase):
    """Gaget leave and remove activity when it becomes private"""
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)
        # Simulate invitation process.
        self.service.create_room('chat@conf.localhost', 'activity1', {})
        self.service.invited('inspector@gadget.localhost',
            'chat@conf.localhost')
        return self.done

    def presence(self, _):
        # Georges's muc presence
        self.send_muc_presence('chat@conf.localhost/georges',
                'georges@foo.org/resource')

        # Gadget's muc presence
        self.send_muc_presence('chat@conf.localhost/inspector')

        reactor.callLater(0, self.later)

    def later(self):
        activity = self.service.model.activity_by_id('activity1')
        assert activity == Activity('activity1', 'chat@conf.localhost', {},
                set(['georges@foo.org']))

        # Georges makes activity private
        message = domish.Element((None, 'message'))
        message['type'] = 'groupchat'
        message['to'] = 'chat@conf.localhost'
        message['from'] = 'chat@conf.localhost/georges'
        properties = message.addElement((ns.ACTIVITY_PROP, 'properties'))
        properties['activity'] = 'activity1'
        properties['room'] = 'chat@conf.localhost'
        property = properties.addElement((None, 'property'))
        property['type'] = 'bool'
        property['name'] = 'private'
        property.addContent('1')
        self.server.send(message)

        reactor.callLater(0, self.later2)

    def later2(self):
        assert not self.model.has_activity('activity1')
        assert len(self.model.activities) == 0
        self.done.callback(None)

class ActivityPropertiesSpoofTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)
        # Simulate invitation process.
        self.service.create_room('chat@conf.localhost', 'activity1', {})
        self.service.invited('inspector@gadget.localhost',
            'chat@conf.localhost')
        return self.done

    def presence(self, _):
        # Bob's muc presence
        self.send_muc_presence('chat@conf.localhost/bob',
            'bob@foo.org/resource')

        # Gadget's muc presence
        self.send_muc_presence('chat@conf.localhost/inspector')

        reactor.callLater(0, self.later)

    def later(self):
        activity = self.model.activity_by_room('chat@conf.localhost')
        assert activity.members == set(['bob@foo.org'])

        activity = self.service.model.activity_by_id('activity1')
        assert activity == Activity('activity1', 'chat@conf.localhost', {},
                set(['bob@foo.org']))

        # bad oscar tries to change properties but is not in the muc
        message = domish.Element((None, 'message'))
        message['type'] = 'groupchat'
        message['to'] = 'anotherchat@conf.localhost'
        message['from'] = 'anotherchat@conf.localhost/oscar'
        properties = message.addElement((ns.ACTIVITY_PROP, 'properties'))
        properties['activity'] = 'activity1'
        properties['room'] = 'chat@conf.localhost'
        property = properties.addElement((None, 'property'))
        property['type'] = 'str'
        property['name'] = 'title'
        property.addContent('Test Activity')
        self.server.send(message)

        reactor.callLater(0, self.later2)

    def later2(self):

        activity = self.service.model.activity_by_id('activity1')
        assert activity == Activity('activity1', 'chat@conf.localhost', {},
                set(['bob@foo.org']))

        self.done.callback(None)

class ViewOfflineTest(TestCase):
    """Test if views are automatically removed when their initator goes
    offline."""
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)
        self.server.addOnetimeObserver('//iq', self.iq2)

        # request a buddy view
        iq = elem_iq(self.server, 'get',
                from_='bob@foo.org', to='gadget.bar.org')(
                    elem(ns.BUDDY, 'view', id='view1')(
                        elem('random', max='10')))
        iq.send()

    def iq(self, stanza):
        # request an activity view
        iq = elem_iq(self.server, 'get',
                    from_='bob@foo.org', to='gadget.bar.org')(
                        elem(ns.ACTIVITY, 'view', id='view1')(
                            elem('random', max='10')))
        iq.send()

    def iq2(self, stanza):
        # Bob goes offline
        presence = domish.Element((ns.CLIENT, 'presence'))
        presence['from'] = 'bob@foo.org'
        presence['to'] = 'gadget.bar.org'
        presence['type'] = 'unavailable'
        self.server.send(presence)

        reactor.callLater(0, self.later)

    def later(self):
        assert len(self.service.activity_views) == 0
        assert len(self.model.activity_views) == 0
        assert len(self.service.buddy_views) == 0
        assert len(self.model.buddy_views) == 0
        self.done.callback(None)

class BuddyCurrentActivityChangeTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.service.model.create_buddy('bob@foo.org', {})

        # Bob changed his current activity. Sending pubsub event
        message = elem('message', from_='bob@foo.org', to='gadget.localhost')(
            elem(ns.PUBSUB_EVENT, 'event')(
                elem('items', node=ns.CURRENT_ACTIVITY)(
                    elem('item')(
                        elem(ns.CURRENT_ACTIVITY, 'activity',
                            room='room1@conf.localhost', type='id1')))))

        self.server.send(message)

        reactor.callLater(0, self.later1)
        return self.done

    def later1(self):
        buddy = self.service.model.buddy_by_id('bob@foo.org')
        assert buddy.current_activity.room, buddy.current_activity.id == ('room1@conf.localhost', 'id1')

        # Bob doesn't have a current activity anymore.
        message = elem('message', from_='bob@foo.org', to='gadget.localhost')(
            elem(ns.PUBSUB_EVENT, 'event')(
                elem('items', node=ns.CURRENT_ACTIVITY)(
                    elem('item')(
                        elem(ns.CURRENT_ACTIVITY, 'activity',
                            room='', type='')))))

        self.server.send(message)
        reactor.callLater(0, self.later2)

    def later2(self):
        buddy = self.service.model.buddy_by_id('bob@foo.org')
        assert buddy.current_activity is None
        self.done.callback(None)

class BuddyNickChangeTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.service.model.create_buddy('bob@foo.org', {})

        # Bob changed his nick. Sending pubsub event
        message = elem('message', from_='bob@foo.org', to='gadget.localhost')(
            elem(ns.PUBSUB_EVENT, 'event')(
                elem('items', node=ns.NICK)(
                    elem('item')(
                        elem(ns.NICK, 'nick')(u'BOB')))))

        self.server.send(message)

        reactor.callLater(0, self.later1)
        return self.done

    def later1(self):
        buddy = self.service.model.buddy_by_id('bob@foo.org')
        assert buddy.alias == 'BOB'

        self.done.callback(None)

class RestBuddyTest(TestCase):
    """Test if buddy is reset and his views closed when he reconnects."""

    def runTest(self):
        self.done = defer.Deferred()
        self.service.model.create_buddy('bob@foo.org', {'color': ('str', 'red')})
        self.server.addOnetimeObserver('//iq', self.iq)

        # request a buddy view
        iq = elem_iq(self.server, 'get',
                from_='bob@foo.org', to='gadget.bar.org')(
                    elem(ns.BUDDY, 'view', id='view1')(
                        elem('random', max='10')))
        iq.send()

    def iq(self, stanza):
        self.server.addOnetimeObserver('//presence', self.presence)

        # Bob is disconnected and then reconnects
        presence = domish.Element((ns.CLIENT, 'presence'))
        presence['from'] = 'bob@foo.org/Resource'
        presence['to'] = 'gadget.foo.org'
        presence['type'] = 'probe'
        self.server.send(presence)

    def presence(self, presence):
        buddy = self.service.model.buddy_by_id('bob@foo.org')
        assert buddy.jid == 'bob@foo.org'
        assert buddy.properties == {}

        assert len(self.service.activity_views) == 0
        assert len(self.model.activity_views) == 0
        assert len(self.service.buddy_views) == 0
        assert len(self.model.buddy_views) == 0
        self.done.callback(None)

class LeaveEmptyRoom(TestCase):
    """Gaget leave and remove activity if there is no one in the room
    when he joined it"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)
        # Simulate invitation process.
        self.service.create_room('chat@conf.localhost', 'activity1', {})
        self.service.invited('inspector@gadget.localhost',
            'chat@conf.localhost')
        return self.done

    def presence(self, _):
        # Gadget's muc presence
        self.send_muc_presence('chat@conf.localhost/inspector')

        reactor.callLater(0, self.later)

    def later(self):
        assert not self.service.model.has_activity('activity1')
        self.done.callback(None)

class MucJoiningError(TestCase):
    """Gadget cannont join the room so removes the activity"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//presence', self.presence)
        # Simulate invitation process.
        self.service.create_room('chat@conf.localhost', 'activity1', {})
        self.service.invited('inspector@gadget.localhost',
            'chat@conf.localhost')
        return self.done

    def presence(self, _):
        # Error when joining
        self.send_muc_presence('chat@conf.localhost/inspector', None, 'error')

        reactor.callLater(0, self.later)

    def later(self):
        assert not self.service.model.has_activity('activity1')

        assert len(self.service.joining_activities) == 0
        assert len(self.service.model.activities) == 0
        self.done.callback(None)
