
import unittest

from gadget.model import GadgetModel

class ViewHandler(object):
    def __init__(self):
        self.activity_found_log = []
        self.activity_gone_log = []
        self.member_added_log = []
        self.member_removed_log = []
        self.buddy_changed_log = []
        self.buddy_current_activity_changed_log = []
        self.buddy_added_log = []
        self.buddy_removed_log = []

    def activity_found(self, activity):
        self.activity_found_log.append(activity)

    def activity_gone(self, activity):
        self.activity_gone_log.append(activity)

    def member_added(self, activity, buddy):
        self.member_added_log.append((activity, buddy))

    def member_removed(self, activity, buddy):
        self.member_removed_log.append((activity, buddy))

    def buddy_changed(self, buddy):
        self.buddy_changed_log.append(buddy)

    def buddy_current_activity_changed(self, buddy):
        self.buddy_current_activity_changed_log.append(buddy)

    def buddy_added(self, buddy):
        self.buddy_added_log.append(buddy)

    def buddy_removed(self, buddy):
        self.buddy_removed_log.append(buddy)

class ModelTest(unittest.TestCase):
    def test_activity_search(self):
        model = GadgetModel()
        assert not model.has_activity('123')
        activity = model.create_activity('123', 'room@conf.foo.org',
                {'type': ('str', 'Paint')})
        assert model.has_activity('123')

        self.assertEquals(activity, model.activity_by_id('123'))
        self.assertRaises(KeyError, model.activity_by_id, '124')

        results = model.activity_by_properties({'type': ('str', 'Paint')})
        self.assertEquals([activity], results)

        results = model.activity_by_properties({'type': ('str', 'Draw')})
        self.assertEquals([], results)

        result = model.activity_by_room('room@conf.foo.org')
        self.assertEquals(activity, result)

        result = model.activity_by_room('room@conf.bar.org')
        self.assertEquals(None, result)

        activity.members.add('alice@foo.org')
        results = model.activity_by_participants(['alice@foo.org'])
        self.assertEquals([activity], results)
        results = model.activity_by_participants(['boo@foo.org'])
        self.assertEquals([], results)

        activity2 = model.create_activity('456', 'room2@conf.foo.org', {})
        activity2.members.add('alice@foo.org')
        activity2.members.add('bob@foo.org')
        results = model.activity_by_participants(['alice@foo.org'])
        self.assertEquals([activity, activity2], sorted(results))
        results = model.activity_by_participants(
            ['alice@foo.org', 'bob@foo.org'])
        self.assertEquals([activity2], results)

        results = model.activity_all()
        self.assertEquals([activity, activity2], sorted(results))

        results = model.activity_search({'room': 'room@conf.foo.org'})
        self.assertEquals([activity], list(results))
        results = model.activity_search({'room': 'room@conf.foo.org', 'buddies':
            ['alice@foo.org']})
        self.assertEquals([activity], list(results))

    def test_buddy_search(self):
        model = GadgetModel()

        assert not model.has_buddy('alice@foo.org')
        self.assertRaises(KeyError, model.buddy_by_id, 'alice@foo.org')
        alice = model.create_buddy('alice@foo.org', {})
        assert model.has_buddy('alice@foo.org')

        result = model.buddy_by_id('alice@foo.org')
        self.assertEquals(alice, result)

        model.buddy_update('alice@foo.org', {'color': ('str', 'blue')})
        results = model.buddy_by_properties({'color': ('str', 'blue')})
        self.assertEquals([alice], results)

        results = model.buddy_all()
        self.assertEquals([alice], results)

        results = model.buddy_search({'jid': 'alice@foo.org'})
        self.assertEquals([alice], list(results))
        results = model.buddy_search({'jid': 'alice@foo.org',
            'properties': {'color': ('str', 'red')}})
        self.assertEquals([], list(results))
        results = model.buddy_search({'jid': 'bob@foo.org'})
        self.assertEquals([], list(results))

        bob = model.create_buddy('bob@foo.org', {'color': ('str', 'red')})
        charles = model.create_buddy('charles@foo.org',
            {'color': ('str', 'red')})
        results = model.buddy_search({'properties': {'color': ('str', 'red')}})
        self.assertEquals([bob, charles], sorted(results))

        results = model.buddy_search({'jid': 'bob@foo.org', 'properties': {'color': ('str', 'red')}})
        self.assertEquals([bob], list(results))

    def test_buddy_random(self):
        model = GadgetModel()
        model.create_buddy('alice@foo.org', {})
        model.create_buddy('bob@foo.org', {})

        result = model.buddy_random(1)
        assert len(result) == 1
        result = model.buddy_random(2)
        assert len(result) == 2
        result = model.buddy_random(3)
        assert len(result) == 2
        self.assertRaises(ValueError, model.buddy_random, 0)
        self.assertRaises(ValueError, model.buddy_random, -1)

    def test_activity_random(self):
        model = GadgetModel()
        model.create_activity('123', 'room1@conf.foo.org', {})
        model.create_activity('456', 'room2@conf.foo.org', {})
        model.create_activity('789', 'room3@conf.foo.org', {})

        result = model.activity_random(2)
        assert len(result) == 2
        result = model.activity_random(5)
        assert len(result) == 3
        self.assertRaises(ValueError, model.activity_random, -1)

    def test_activity_props_view(self):
        model = GadgetModel()
        handler = ViewHandler()
        activity1 = model.create_activity(123, 'room1@conf.foo.org',
            {'color': ('str', 'red')})
        view = model.activity_view(handler, 3, {'properties': {'color': ('str', 'red')}})

        self.assertEquals([activity1], view.results)
        self.assertEquals([], handler.activity_found_log)
        self.assertEquals([], handler.activity_gone_log)

        # buddy1 is member of activity1
        buddy1 = model.create_buddy('buddy1@localhost', {})
        activity1.members.add('buddy1@localhost')

        # buddy1 changed his properties
        model.buddy_update(buddy1.jid, {'color': ('str', 'red')})
        self.assertEquals([buddy1], handler.buddy_changed_log)

        # buddy1 changed his current-activity
        model.buddy_change_current_activity(buddy1.jid, ('room1', 'id1'))
        self.assertEquals([buddy1], handler.buddy_current_activity_changed_log)

        # Non-matching activity added to model.
        activity2 = model.create_activity(456, 'room2@conf.foo.org', {})

        self.assertEquals([activity1], view.results)
        self.assertEquals([], handler.activity_found_log)
        self.assertEquals([], handler.activity_gone_log)

        # Matching activities added to model.
        activity3 = model.create_activity(768, 'room3@conf.foo.org',
            {'color': ('str', 'red')})
        activity4 = model.create_activity(901, 'room4@conf.foo.org',
            {'color': ('str', 'red')})

        self.assertEquals([activity1, activity3, activity4], view.results)
        self.assertEquals([activity3, activity4], handler.activity_found_log)
        self.assertEquals([], handler.activity_gone_log)

        # Matching activity added to the model but not added as the view is
        # full
        activity5 = model.create_activity(1000, 'room4@conf.foo.org',
            {'color': ('str', 'red')})

        self.assertEquals([activity1, activity3, activity4], view.results)
        self.assertEquals([activity3, activity4], handler.activity_found_log)
        self.assertEquals([], handler.activity_gone_log)

        # Matching activity removed from model.
        # activity5 is added now
        model.activity_remove(activity3.id)

        self.assertEquals([activity1, activity4, activity5], view.results)
        self.assertEquals([activity3, activity4, activity5], handler.activity_found_log)
        self.assertEquals([activity3], handler.activity_gone_log)

        # Non-matching activity removed from model.
        model.activity_remove(activity2.id)

        self.assertEquals([activity1, activity4, activity5], view.results)
        self.assertEquals([activity3, activity4, activity5], handler.activity_found_log)
        self.assertEquals([activity3], handler.activity_gone_log)

        # Matching activity stops matching.
        model.activity_update(activity1.id, {'color': ('str', 'blue')})

        self.assertEquals([activity4, activity5], view.results)
        self.assertEquals([activity3, activity4, activity5], handler.activity_found_log)
        self.assertEquals([activity3, activity1], handler.activity_gone_log)

        # Non-matching activity starts matching.
        model.activity_update(activity1.id, {'color': ('str', 'red')})

        self.assertEquals([activity4, activity5, activity1], view.results)
        self.assertEquals([activity3, activity4, activity5, activity1], handler.activity_found_log)
        self.assertEquals([activity3, activity1], handler.activity_gone_log)

        # add a member to activity1
        model.activity_add_member(activity1.id, 'user1@localhost')
        self.assertEquals([(activity1, 'user1@localhost')], handler.member_added_log)

        # remove a member from activity1
        model.activity_remove_member(activity1.id, 'user1@localhost')
        self.assertEquals([(activity1, 'user1@localhost')], handler.member_removed_log)

        # Deleted views are automatically removed from the model
        assert len(model.activity_views) == 1
        del view
        assert len(model.activity_views) == 0

    def test_buddy_props_view(self):
        model = GadgetModel()
        handler = ViewHandler()
        buddy1 = model.create_buddy('buddy1@localhost',
            {'color': ('str', 'red')})
        view = model.buddy_view(handler, 3, {'properties': {'color': ('str', 'red')}})

        self.assertEquals([buddy1], view.results)
        self.assertEquals([], handler.buddy_added_log)
        self.assertEquals([], handler.buddy_removed_log)
        self.assertEquals([], handler.buddy_current_activity_changed_log)

        # Non-matching buddies added to model.
        buddy2 = model.create_buddy('buddy2@localhost', {})
        buddy5 = model.create_buddy('buddy5@localhost', {})

        self.assertEquals([buddy1], view.results)
        self.assertEquals([], handler.buddy_added_log)
        self.assertEquals([], handler.buddy_removed_log)
        self.assertEquals([], handler.buddy_current_activity_changed_log)

        # Matching buddies added to model.
        buddy3 = model.create_buddy('buddy3@localhost',
            {'color': ('str', 'red')})
        buddy4 = model.create_buddy('buddy4@localhost',
            {'color': ('str', 'red')})

        self.assertEquals([buddy1, buddy3, buddy4], view.results)
        self.assertEquals([buddy3, buddy4], handler.buddy_added_log)
        self.assertEquals([], handler.buddy_removed_log)

        # Matching activity added to the model but not added as the view is
        # full
        buddy5 = model.create_buddy('buddy5@localhost',
            {'color': ('str', 'red')})

        self.assertEquals([buddy1, buddy3, buddy4], view.results)
        self.assertEquals([buddy3, buddy4], handler.buddy_added_log)
        self.assertEquals([], handler.buddy_removed_log)
        self.assertEquals([], handler.buddy_current_activity_changed_log)

        # Matching buddy removed from model. buddy5 is added now.
        model.buddy_remove(buddy3.jid)

        self.assertEquals([buddy1, buddy4, buddy5], view.results)
        self.assertEquals([buddy3, buddy4, buddy5], handler.buddy_added_log)
        self.assertEquals([buddy3], handler.buddy_removed_log)
        self.assertEquals([], handler.buddy_current_activity_changed_log)

        # Non-matching buddy removed from model.
        model.buddy_remove(buddy2.jid)

        self.assertEquals([buddy1, buddy4, buddy5], view.results)
        self.assertEquals([buddy3, buddy4, buddy5], handler.buddy_added_log)
        self.assertEquals([buddy3], handler.buddy_removed_log)
        self.assertEquals([], handler.buddy_current_activity_changed_log)

        # Matching activity added to the model but not added as the view is
        # full
        buddy6 = model.create_buddy('buddy6@localhost',
            {'color': ('str', 'red')})

        # Matching buddy stops matching. buddy6 is added now.
        model.buddy_update(buddy1.jid, {'color': ('str', 'blue')})

        self.assertEquals([buddy4, buddy5, buddy6], view.results)
        self.assertEquals([buddy3, buddy4, buddy5, buddy6], handler.buddy_added_log)
        self.assertEquals([buddy3, buddy1], handler.buddy_removed_log)
        self.assertEquals([], handler.buddy_current_activity_changed_log)

        # Remove buddy6
        model.buddy_remove(buddy6.jid)

        # Non-matching buddy starts matching.
        model.buddy_update(buddy1.jid, {'color': ('str', 'red')})

        self.assertEquals([buddy4, buddy5, buddy1], view.results)
        self.assertEquals([buddy3, buddy4, buddy5, buddy6, buddy1], handler.buddy_added_log)
        self.assertEquals([buddy3, buddy1, buddy6], handler.buddy_removed_log)
        self.assertEquals([], handler.buddy_current_activity_changed_log)

        # Matching buddy update properties
        model.buddy_update(buddy1.jid, {'name': ('str', 'foo'),
            'color': ('str', 'red')})
        self.assertEquals([buddy4, buddy5, buddy1], view.results)
        self.assertEquals([buddy1], handler.buddy_changed_log)
        self.assertEquals([], handler.buddy_current_activity_changed_log)

        # Non-matching buddy update properties
        model.buddy_update(buddy5.jid, {'name': ('str', 'foo')})
        self.assertEquals([buddy4, buddy1], view.results)
        self.assertEquals([buddy1], handler.buddy_changed_log)
        self.assertEquals([], handler.buddy_current_activity_changed_log)

        # Matching buddy change his current activity
        model.buddy_change_current_activity(buddy1.jid, ('room1', 'id1'))
        self.assertEquals([buddy4, buddy1], view.results)
        self.assertEquals([buddy1], handler.buddy_changed_log)
        self.assertEquals([buddy1], handler.buddy_current_activity_changed_log)

        # Non-matching buddy change his current activity
        model.buddy_change_current_activity(buddy5.jid, ('room2', 'id2'))
        self.assertEquals([buddy4, buddy1], view.results)
        self.assertEquals([buddy1], handler.buddy_changed_log)
        self.assertEquals([buddy1], handler.buddy_current_activity_changed_log)

        # Deleted views are automatically removed from the model
        assert len(model.buddy_views) == 1
        del view
        assert len(model.buddy_views) == 0

    def test_buddy_alias_view(self):
        model = GadgetModel()
        handler = ViewHandler()
        buddy1 = model.create_buddy('buddy1@localhost', {})
        buddy1.alias = 'tom'

        view = model.buddy_view(handler, 3, {'alias': 'tom'})

        self.assertEquals([buddy1], view.results)
        self.assertEquals([], handler.buddy_added_log)

        # Non-matching buddy added to model.
        buddy2 = model.create_buddy('buddy2@localhost', {})
        buddy5 = model.create_buddy('buddy5@localhost', {})

        self.assertEquals([buddy1], view.results)
        self.assertEquals([], handler.buddy_added_log)

        # Matching buddies added to model.
        buddy3 = model.create_buddy('buddy3@localhost',
            {'color': ('str', 'red')})
        model.buddy_change_alias(buddy3.jid, 'tom')

        self.assertEquals([buddy1, buddy3], view.results)
        self.assertEquals([buddy3], handler.buddy_added_log)

        # Matching buddy removed from model.
        model.buddy_remove(buddy3.jid)

        self.assertEquals([buddy1], view.results)
        self.assertEquals([buddy3], handler.buddy_added_log)
        self.assertEquals([buddy3], handler.buddy_removed_log)

        # Non-matching buddy removed from model.
        model.buddy_remove(buddy2.jid)

        self.assertEquals([buddy1], view.results)
        self.assertEquals([buddy3], handler.buddy_added_log)
        self.assertEquals([buddy3], handler.buddy_removed_log)

        # Matching buddy stops matching.
        model.buddy_change_alias(buddy1.jid, 'alex')

        self.assertEquals([], view.results)
        self.assertEquals([buddy3], handler.buddy_added_log)
        self.assertEquals([buddy3, buddy1], handler.buddy_removed_log)

        # Non-matching buddy starts matching.
        model.buddy_change_alias(buddy5.jid, 'tom')

        self.assertEquals([buddy5], view.results)
        self.assertEquals([buddy3, buddy5], handler.buddy_added_log)

        # Deleted views are automatically removed from the model
        assert len(model.buddy_views) == 1
        del view
        assert len(model.buddy_views) == 0

    def test_activity_participants_view(self):
        model = GadgetModel()
        handler = ViewHandler()
        activity1 = model.create_activity(1, 'room1@conf.foo.org', {})
        activity1.members.add('alice@localhost')
        activity1.members.add('bob@localhost')
        activity2 = model.create_activity(2, 'room2@conf.foo.org', {})
        activity2.members.add('alice@localhost')

        # ask a view containing activities in which Alice and Bob are
        # participating
        view = model.activity_view(handler, 3, {'participants': ["alice@localhost", "bob@localhost"]})

        self.assertEquals([activity1], view.results)
        self.assertEquals([], handler.activity_found_log)
        self.assertEquals([], handler.activity_gone_log)

        # Bob joins activity2
        model.activity_add_member(2, 'bob@localhost')

        self.assertEquals([activity1, activity2], view.results)
        self.assertEquals([activity2], handler.activity_found_log)
        self.assertEquals([], handler.activity_gone_log)

        # Alice left activity1
        model.activity_remove_member(1, 'alice@localhost')

        self.assertEquals([activity2], view.results)
        self.assertEquals([activity2], handler.activity_found_log)
        self.assertEquals([activity1], handler.activity_gone_log)

    def test_activity_participants_and_properties_view(self):
        model = GadgetModel()
        handler = ViewHandler()
        activity1 = model.create_activity(1, 'room1@conf.foo.org',
            {'type': ('str', 'Connect')})
        activity1.members.add('alice@localhost')
        activity2 = model.create_activity(2, 'room2@conf.foo.org', {})
        activity2.members.add('alice@localhost')
        activity3 = model.create_activity(3, 'room3@conf.foo.org',
            {'type': ('str', 'Connect')})
        activity3.members.add('bob@localhost')

        # ask a view containing Connect activities in which Alice is
        # participating
        view = model.activity_view(handler, 3, {'properties': {'type': ('str', 'Connect')},
            'participants': ["alice@localhost"]})

        self.assertEquals([activity1], view.results)
        self.assertEquals([], handler.activity_found_log)
        self.assertEquals([], handler.activity_gone_log)

        # Alice joined activity3
        model.activity_add_member(3, 'alice@localhost')

        self.assertEquals([activity1, activity3], view.results)
        self.assertEquals([activity3], handler.activity_found_log)
        self.assertEquals([], handler.activity_gone_log)

        # activity2 became a Connect activity
        model.activity_update(2, {'type': ('str', 'Connect')})

        self.assertEquals([activity1, activity3, activity2], view.results)
        self.assertEquals([activity3, activity2], handler.activity_found_log)
        self.assertEquals([], handler.activity_gone_log)

        # Alice left activity1
        model.activity_remove_member(1, 'alice@localhost')

        self.assertEquals([activity3, activity2], view.results)
        self.assertEquals([activity3, activity2], handler.activity_found_log)
        self.assertEquals([activity1], handler.activity_gone_log)
